package cn.dankal.test.order.service.impl;

import cn.dankal.test.client.StorageServiceClient;
import cn.dankal.test.common.result.Result;
import cn.dankal.test.order.model.Order;
import cn.dankal.test.order.repository.OrderMapper;
import cn.dankal.test.order.service.OrderService;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import io.seata.core.context.RootContext;
import io.seata.core.exception.TransactionException;
import io.seata.spring.annotation.GlobalTransactional;
import io.seata.tm.api.GlobalTransactionContext;
import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.util.Random;

/**
 * {@link OrderService} {@link ServiceImpl}
 * OrderServiceImpl
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2020/10/22
 */
@Service
public class OrderServiceImpl extends ServiceImpl<OrderMapper, Order> implements OrderService {

    /**
     * StorageServiceClient
     */
    private final StorageServiceClient storageServiceClient;

    public OrderServiceImpl(StorageServiceClient storageServiceClient) {
        this.storageServiceClient = storageServiceClient;
    }

    @GlobalTransactional(rollbackFor = {Exception.class})
    @Override
    public boolean placeOrder(Order order) throws TransactionException {
        // 参数校验
        if (null == order) {
            return false;
        }
        String commodityCode = order.getCommodityCode();
        Integer count = order.getCount();
        if (count <= 0) {
            return false;
        }
        if (StringUtils.isBlank(commodityCode)) {
            return false;
        }
        String userId = order.getUserId();
        if (StringUtils.isBlank(userId)) {
            return false;
        }

        // 扣减库存
        Boolean deduct = storageServiceClient.deduct(commodityCode, count);
        if (deduct == null || !deduct) {
            // 库存扣减失败 事务回滚
            GlobalTransactionContext.reload(RootContext.getXID()).rollback();
            return false;
        }
        // 设置订单金额
        order.setMoney(BigDecimal.valueOf(new Random().nextDouble()*1000));
        // TODO 关于 Seata 对于 SELECT 语句的代理
        // @see  http://seata.io/zh-cn/docs/overview/what-is-seata.html --> 读隔离
        // 查询用户余额
        // accountService.getOne(new LambdaQueryWrapper<Account>().eq(Account::getUserId,userId).last("FOR UPDATE"));
        // 创建订单
        return this.save(order);
    }

    @GlobalTransactional(rollbackFor = Exception.class)
    @Override
    public Boolean placeOrderException() {
        Order order = new Order()
                .setMoney(BigDecimal.valueOf(-1))
                .setUserId("nil")
                .setCount(-1)
                .setCommodityCode("nil");
        this.save(order);
        // 全局异常捕获器位于全局事务的内层
        //  storage-server --> cn.dankal.test.storage.advice.ExceptionAdvice.handlerException 已经将异常处理
        // 所以 seata 在此处无法感知异常的发生,所以事务是正常提交的
        Result<Object> execResult = storageServiceClient.deductWithException();
        System.out.println("deductWithException --> result = " + execResult);
        return false;
    }

    @GlobalTransactional(rollbackFor = Exception.class)
    @Override
    public Boolean placeOrderExceptionRollbackByApi() throws TransactionException {
        Order order = new Order()
                .setMoney(BigDecimal.valueOf(-1))
                .setUserId("nil")
                .setCount(-1)
                .setCommodityCode("nil");
        this.save(order);
        Result<Object> execResult = storageServiceClient.deductWithException();
        if (!execResult.getIsSuccess()) {
            // 操作异常 手动混滚
            GlobalTransactionContext.reload(RootContext.getXID()).rollback();
            // 直接返回 当前事务已经回滚不可再执行任何事务型操作
            return  false;
        }
        // TODO  etc
        // this.update(null);
        return true;
    }
}
